{-
	Denotational Semantics
	A block-structured language

	!!! NOT COMPLETED !!!

Abstract Syntax:

P of Program
K of Block
D of Declaration
C of Command
E of Expression
B of Boolean-expr
I of Identifier
N of Numeral

P ::= K '.'
K ::= 'begin' D ';' C 'end'
D ::= D1 ';' D2 | 'const' I '=' N | 'var' I
C ::= C1 ';' C2 | I ':=' E | 'while' B 'do' C | K
E ::= E1 '+' E2 | I | N
B ::= E1 '=' E2 | 'not' B (ML)

Semantic algebras:

I.	Truth Values
	Domain t of Tr = Boolean
	Operations
	 true, false : Tr
	 not : Tr -> Tr

II.	Identifiers
	Domain i of Id = Identifier (String)

III.	Natural Numbers
	Domain n of Nat = Int
	Operations
	 zero, one, ... : Nat
	 plus : Nat x Nat -> Nat
	 equals : Nat x Nat -> Tr

IV.	Expressible values
	Domain x of Expressible-value = Nat + Errvalue
		    where Errvalue = Unit

V.	Storage locations
	Domain l of Location
	Operations
	 first-locn : Location
	 next-locn : Location -> Location
	 equal-locn : Location -> Location -> Tr
	 lessthan-locn : Location -> Location -> Tr

VI.	Denotable values
	Domain d of Denotable-value = Location + Nat + Errvalue
	            where Errvalue = Unit

VII.	Environment: a map to denotable values and the maximum store location
	Domain e of Environment = (Id -> Denotable-value) x Location
	Operations
	 emptyenv : Location -> Environment
	  emptyenv = \l.((\i.inErrvalue()),l)

	 accessenv : Id -> Environment -> Denotable-value
	  accessenv = \i.\(map,l).map(i)

	 updateenv : Id -> Denotable-value -> Environment -> Environment
	  updateenv = \i.\d.\(map,l).([i=>d]map,l)

	reserve-locn : Environment -> (Location x Environment)
	 reserve-locn = \(map,l).(l,(map,next-locn(l)))

VIII.	Storeable values
	Domain v of Storeable-value = Nat

IX.	Stores
	Domain s of Store = Location -> Storeable-value
	Operations
	 access : Location -> Store -> Storeable-value
	  access = \l.\s.s(i)

	 update : Location -> Storeable-value -> Store -> Store 
	  update = \l.\v.\s.[i=>n]s

X.	Run-time store, labeled with status of computation
	Domain p of Poststore = OK + Err
	            where OK = Err = Store
	Operations
	 return : Store -> Poststore
	  return = \s.inOK(s)

	 signalerr : Store -> Poststore
	  signalerr = \s.inErr(s)

	 check : (Store -> Poststore_) -> (Poststore_ -> Poststore_)
	  check f = \_p.cases p of
			isOK(s) -> (f s)
			isErr(s) -> p end

Valuation functions:

PP : Program -> Location -> Store -> Poststore_

  PP[K '.'] = \l.KK[K](emptyenv l)

KK : Block -> Environment -> Store -> Poststore_

  KK['begin' D ';' C 'end'] = \e.CC[C](DD[D]e)

DD : Declaration -> Environment -> Environment

  DD[D1 ';' D2] = DD[D2] * DD[D1] ( ==> DD[D2](DD[D1]) )
  DD['const' I '=' N] = updateenv [I] inNat(NN[N])
  DD['var' I] = \e.let (l',e') = (reserve-locn e) 
				in (updateenv [I] inLocation(l') e')

CC : Command -> Environment -> Store -> Poststore_

  CC[C1 ';' C2] = \e.(check (CC[C2]e)) * (CC[C1]e)
  CC[I ':=' E] = \e.\s.cases (accessenv [I] e) of
			isLocation(l) -> (cases (EE[E]e s) of
					   isNat(n)->(return (update l n s))
					   isErrvalue() -> (signalerr s) end)
			isNat(n) -> (signalerr s)
			isErrvalue() -> (signalerr s) end
  CC['while' B 'do' C] = \e.fix(\f.\s.cases (BB[B]e s) of
					isTr(t)->(t->(check f) * (CC[C]e)
                                                     return )( s )
					isErrvalue() -> (signalerr s) end)
  CC[K] = KK[K]

EE : Expression -> Environment -> Store -> Expressible-value

  EE[E1 '+' E2] = \e.\s.cases (EE[E1]e s) of
			  isNat(n1) -> (cases (EE[E2]e s) of
					 isNat(n2) -> inNat(n1 plus n2)
					 isErrvalue() -> inErrvalue() end)
			  isErrvalue() -> inErrvalue() end
  EE[I] = \e.\s.cases (accessenv [I] e) of
		  isLocation(l) -> inNat(access l s)
		  isNat(n) -> inNat(n)
		  isErrvalue() -> inErrvalue() end
  EE[N] = \e.\s.inNat(NN[N])

BB : Boolean-expr -> Environment -> Store -> (Tr + Errvalue)

  BB[E1 '=' E1] = \e.\s.cases (EE[E1]e s) of
			isNat(n1) -> (cases (EE[E2]e s) of
					isNat(n2) -> inTr(n1 equals n2)
					inErrvalue() -> inErrvalue() end)
			isErrvalue() -> inErrvalue() end

  BB['not' B] = \e.\s.cases (BB[B]e s) of
			isTr(t) -> inTr(t)
			isErrvalue() -> inErrvalue() end

NN : Numeral -> Nat

-}

infixr 0 !!!
infixr 1 &&&, &&>

------------------------------------------------------------------------------

{-
P ::= K '.'
K ::= 'begin' D ';' C 'end'
D ::= D1 ';' D2 | 'const' I '=' N | 'var' I
C ::= C1 ';' C2 | I ':=' E | 'while' B 'do' C | K
E ::= E1 '+' E2 | I | N
B ::= E1 '=' E2 | 'not' B (ML)
-}

data Program =
	  Dot Block
	| Err

data Block =
	  Begin Declaration Command

data Declaration =
	  Seq Declaration Declaration
	| Const Identifier Numeral
	| Var Identifier

data Command =
	  CSeq Command Command
	| Assign Identifier Expression
	| While BooleanExpr Command
	| NewBlock Block

data Expression =
	  Plus Expression Expression
	| Id Identifier
	| Num Numeral

data BooleanExpr =
	  Equal Expression Expression
	| Not BooleanExpr

data Identifier =
	  Ident String

data Numeral = 
	  Number String

-------------------------------------------------------------------------------
---- Definitions for syntax analysis ------------------------------------------
-------------------------------------------------------------------------------

data Terminal = 
	  TDot
	| TBegin
	| TEnd
	| TSemicolon
	| TEqual
	| TConst
	| TVar
	| TAssign
	| TWhile
	| TDo
	| TPlus
	| TNot
	| TName
	| TNumber
	| TEOF

instance Eq Terminal where
	TDot == TDot = True
	TBegin == TBegin = True
	TEnd == TEnd = True
	TSemicolon == TSemicolon = True
	TEqual == TEqual = True
	TConst == TConst = True
	TVar == TVar = True
	TAssign == TAssign = True
	TWhile == TWhile = True
	TDo == TDo = True
	TPlus == TPlus = True
	TNot == TNot = True
	TName == TName = True
	TNumber == TNumber = True
	_ == _ = False

---- Scanner -----------------------------------------------------------------

getNextToken :: String -> ( Terminal, String, String )

getNextToken [] = ( TEOF, "", [] )
getNextToken (';':cs) = ( TSemicolon, "", cs )
getNextToken ('.':cs) = ( TDot, "", cs )
getNextToken (':':'=':cs) = ( TAssign, "", cs )
getNextToken ('+':cs) = ( TPlus, "", cs )
getNextToken ('=':cs) = ( TEqual, "", cs )

getNextToken (c:cs) | isSpace c = getNextToken cs
                    | isDigit c = ( TNumber, (c:ns), cs' )
                    | isAlpha c = ( (getKey (c:ns')), (c:ns'), cs'' )
                    | otherwise = getNextToken cs
                    where (ns, cs' )  = ( getWhile isDigit cs ,
                                          getRemain isDigit cs )
                          (ns', cs'') = ( getWhile isAlphanum cs,
                                          getRemain isAlphanum cs )

getKey :: String -> Terminal

getKey "begin" = TBegin
getKey "end" = TEnd
getKey "const" = TConst
getKey "var" = TVar
getKey "while" = TWhile
getKey "do" = TDo
getKey "not" = TNot
getKey _ = TName 

getWhile :: (Char -> Bool) -> String -> String
getWhile _ [] = []
getWhile p (c:cs) = if p c
                    then c : getWhile p cs
                    else []

getRemain :: (Char -> Bool) -> String -> String
getRemain _ [] = []
getRemain p css@(c:cs) = if p c
                         then getRemain p cs
                         else css

---- Parser -------------------------------------------------------------------

type ParserState = String

data ParserResult a =
	  Succ a ParserState
	| Fail

type Parser a = ParserState -> ParserResult a


epsilon :: a -> Parser a
epsilon a = \ps -> Succ a ps


literal :: Terminal -> Parser ()
literal t = \s -> let (nt, v, sr) = getNextToken s in
                            if nt == t
                            then Succ () sr
                            else Fail 

value :: Terminal -> Parser String
value t = \s -> let (nt, v, sr) = getNextToken s in
                            if nt == t
                            then Succ v sr
                            else Fail

(!!!) :: Parser a -> Parser a -> Parser a

p1 !!! p2 = \s -> case p1 s of
			result@(Succ _ _) -> result
			Fail -> p2 s

(&&&) :: Parser a -> Parser b -> Parser b

p1 &&& p2 = \s -> case p1 s of
			Succ v1 s1 -> case p2 s1 of
					result@(Succ _ _) -> result
					Fail -> Fail
			Fail -> Fail

(&&>) :: Parser a -> (a -> Parser b) -> Parser b

p1 &&> p2 = \s -> case p1 s of
			Succ v1 s1 -> case p2 v1 s1 of
					result@(Succ _ _) -> result
					Fail -> Fail
			Fail -> Fail

---- language parser --------------------------------------------------------

parse_N :: Parser Numeral

parse_N =
	value TNumber &&> (\n -> epsilon (Number n) )

parse_I :: Parser Identifier

parse_I =
	value TName &&> (\i -> epsilon (Ident i ) )

parse_B :: Parser BooleanExpr

parse_B =
	parse_E &&> (\e1 -> literal TEqual &&& parse_E &&>
			(\e2 -> epsilon(Equal e1 e2) ) )

	!!!

	literal TNot &&& parse_B &&> (\be -> epsilon (Not be) )

parse_E :: Parser Expression

parse_E =
	parse_E' &&> (\e1 -> literal TPlus &&& parse_E &&>
			(\e2 -> epsilon (Plus e1 e2) ) )

	!!!

	parse_E'
	
parse_E' :: Parser Expression

parse_E' =
	parse_I &&> (\id -> epsilon (Id id) )

	!!!

	parse_N &&> (\n -> epsilon (Num n) )

parse_C :: Parser Command

parse_C =
	parse_C' &&> (\c1 -> literal TSemicolon &&& parse_C &&>
			(\c2 -> epsilon (CSeq c1 c2) ) )

	!!!

	parse_C'

parse_C' :: Parser Command

parse_C' =
	parse_I &&> (\i -> literal TAssign &&& parse_E &&>
			(\e -> epsilon(Assign i e) ) )

	!!!

	literal TWhile &&& parse_B &&> 
		(\be -> literal TDo &&&
			parse_C &&> (\c -> epsilon(While be c) ) )

	!!!

	parse_K &&> (\b -> epsilon(NewBlock b) )

parse_D :: Parser Declaration

parse_D =
	parse_D' &&> (\d1 -> literal TSemicolon &&& parse_D &&>
			(\d2 -> epsilon(Seq d1 d2) ) )

	!!!

	parse_D'

parse_D' :: Parser Declaration

parse_D' =
	literal TConst &&& parse_I &&>
		(\i -> literal TEqual &&& parse_N &&>
		  (\n -> epsilon(Const i n) ) )

	!!!

	literal TVar &&& parse_I &&> (\i -> epsilon(Var i) )

parse_K :: Parser Block

parse_K =
	literal TBegin &&& parse_D &&> 
		(\d -> literal TSemicolon &&& parse_C &&>
		  (\c -> literal TEnd &&& epsilon(Begin d c) ) )

parse_P :: Parser Program

parse_P =
	parse_K &&> (\b -> literal TDot &&& epsilon (Dot b) )

------------------------------------------------------------------------------
---- Interpreter -------------------------------------------------------------
------------------------------------------------------------------------------

interpreter :: String -> Program

interpreter = \s -> case parse_P s of
			Succ p _ -> p
			Fail -> Err


