{-
	Denotational Semantics
	The PICT-language

-------------------------------------------------------------------------------
--- Concrete syntax:                                                        ---
-------------------------------------------------------------------------------

Program ::=	(add-on ML)
	  DecList

DecList ::=	(add-on ML)
	  Dec
	| Dec DecList

Proc ::=  
	  Val '?' Abs
	| Val '?*' Abs
	| PProc

PProc ::=
	  PProc '|' AProc
	| AProc

AProc ::=
	  Val '!' Val
	| 'let' Dec 'in' Proc 'end'
	| '(' Proc ')'

Dec  ::=
	  'new' Id
	| 'new' Id ':' Type
	| 'type' Id '=' Type
	| 'run' Proc	(add-on ML)

Abs ::=
	  Pat '>' Proc

Val ::=
	  Id
	| '[' ']'
	| '[' ValList ']'
	| 'record' 'end'
	| Val 'with' Id '=' Val 'end'
	| '(' Val ')'

ValList ::=	(add-on ML)
	  Val
	| Val ',' ValList

Pat ::=
	  Id
	| Id ':' Type
	| '_'
	| '_' ':' Type
	| Id '\ttbox{@}' Pat
	| Id ':' Type '\ttbox{@}' Pat
	| '[' ']'
	| '[' PatList ']'
	| 'record' 'end'
	| 'record' FieldList 'end'

PatList ::=	(add-on ML)
	  Pat
	| Pat ',' PatList

FieldList ::=
	  Id '=' Pat
	| Id '=' Pat ',' FieldList

Type ::=
	  '^' AType
	| '!' Type
	| '?' Type
	| AType

AType ::=
	  '[' ']'
	| '[' TypeList ']'
	| 'record' 'end'
	| Type 'with' Id ':' Type 'end'
	| 'Top'
	| '(' Type ')'

TypeList ::=
	  Type
	| Type ',' TypeList

-------------------------------------------------------------------------------
--- Abstract syntax:                                                        ---
-------------------------------------------------------------------------------

Program ::=	(add-on ML)
	  DecList

DecList ::=	(add-on ML)
	  Dec
	| Dec DecList

Proc ::=
	  Val '?' Abs
	| Val '?*' Abs
	| Val '!' Val
	| Proc '|' Proc
	| 'let' Dec 'in' Proc 'end

Dec ::=
	  'new' Id
	| 'new' Id ':' Type
	| 'type' Id '=' Type
	| 'run' Proc	(add-on ML)

Abs ::=
	  Pat '>' Proc

Pat ::=
	  Id
	| Id ':' Type
	| '[' ']'
	| '[' PatList ']'
	| '{' '}'
	| '{' FieldList '}'
	| Id '@' Pat
	| Id ':' Type '@' Pat
	| '_'
	| '_' ':' Type

PatList ::=	(add-on ML)
	  Pat
	| Pat ',' PatList

FieldList ::=	(add-on ML)
	  Id '=' Pat
	| Id '=' Pat ',' FieldList

Val ::=
	  Id
	| '[' ']'
	| '[' ValList ']'
	| 'record' 'end'
	| Val 'with' Id '=' Val 'end'	

ValList ::=	(add-on)
	  Val
	| Val ',' ValList

Type ::=
	  '^' Type
	| '!' Type
	| '?' Type
	| '[' ']'
	| '[' TypeList ']'
	| 'Record' 'end'
	| Type 'with' Id ':' Type 'end'
	| 'Top'

TypeList ::=
	  Type
	| Type ',' TypeList

-}

infixr 0 !!!
infixr 1 &&&, &&>

-------------------------------------------------------------------------------
--- Abstract syntax definition ------------------------------------------------
-------------------------------------------------------------------------------

data Program =
	  Decls [Declaration]

data Declaration =
	  NewDecl Ident Type
	| TypeDecl Ident Type
	| RunDecl Process

data Process =
	  Input Value Abstraction
	| RepeatedIn Value Abstraction
	| Output Value Value
	| Composition Process Process
	| Let Declaration Process

data Abstraction =
	  Abs Pattern Process

data Pattern =
	  PatternId Ident Type
	| PatternList [Pattern]
	| Fields [(Ident,Pattern)]
	| Ampersand Ident Type Pattern
	| Any Type

data Value =
	  ValId Ident
	| ValPattern [Value]
	| ValRecord
	| ValWith Value Ident Value

data Type =
	  Channel Type
	| OutChan Type
	| InChan Type
	| TypePattern [Type]
	| TypeRecord
	| TypeWith Type Ident Type
	| Top
	| NoType

data Ident =
	  Identifier String

-------------------------------------------------------------------------------
--- Syntax analysis -----------------------------------------------------------
-------------------------------------------------------------------------------

data Terminal =
	  TInput
	| TRepInput
	| TComposition
	| TOutput
	| TLET
	| TIN
	| TEND
	| TLpar
	| TRpar
	| TNEW
	| TColon
	| TTYPE
	| TEqual
	| TRUN
	| TAbs	-- '>'
	| TLb
	| TRb
	| TComma
	| TRECORD
	| TWITH
	| TAny
	| TAmpersand
	| TChan
	| TBIGRECORD
	| TTOP
	| TName
	| TEOF

instance Eq Terminal where
	TInput == TInput = True
	TRepInput == TRepInput = True
	TComposition == TComposition = True
	TOutput == TOutput = True
	TLET == TLET = True
	TIN == TIN = True
	TEND == TEND = True
	TLpar == TLpar = True
	TRpar == TRpar = True
	TNEW == TNEW = True
	TColon == TColon = True
	TTYPE == TTYPE = True
	TEqual == TEqual = True
	TRUN == TRUN = True
	TAbs == TAbs = True
	TLb == TLb = True
	TRb == TRb = True
	TComma == TComma = True
	TRECORD == TRECORD = True
	TWITH == TWITH = True
	TAny == TAny = True
	TAmpersand == TAmpersand = True
	TChan == TChan = True
	TBIGRECORD == TBIGRECORD = True
	TTOP == TTOP = True
	TName == TName = True
	_ == _ = False

---- Scanner -----------------------------------------------------------------

getNextToken :: String -> ( Terminal, String, String )

getNextToken [] = ( TEOF, "", [] )
getNextToken ('?':'*':cs) = ( TRepInput, "", cs )
getNextToken ('?':cs) = ( TInput, "", cs )
getNextToken ('|':'=':cs) = ( TComposition, "", cs )
getNextToken ('!':cs) = ( TOutput, "", cs )
getNextToken ('(':cs) = ( TLpar, "", cs )
getNextToken (')':cs) = ( TRpar, "", cs )
getNextToken (':':cs) = ( TColon, "", cs )
getNextToken ('=':cs) = ( TEqual, "", cs )
getNextToken ('>':cs) = ( TAbs, "", cs )
getNextToken ('[':cs) = ( TLb, "", cs )
getNextToken (']':cs) = ( TRb, "", cs )
getNextToken (',':cs) = ( TComma, "", cs )
getNextToken ('_':cs) = ( TAny, "", cs )
getNextToken ('@':cs) = ( TAmpersand, "", cs )
getNextToken ('^':cs) = ( TChan, "", cs ) 

getNextToken (c:cs) | isSpace c = getNextToken cs
                    | isAlpha c = ( (getKey (c:ns')), (c:ns'), cs'' )
                    | otherwise = getNextToken cs
                    where (ns, cs' )  = ( getWhile isDigit cs ,
                                          getRemain isDigit cs )
                          (ns', cs'') = ( getWhile isAlphanum cs,
                                          getRemain isAlphanum cs )

getKey :: String -> Terminal

getKey "let" = TLET
getKey "in" = TIN
getKey "end" = TEND
getKey "new" = TNEW
getKey "type" = TTYPE
getKey "run" = TRUN
getKey "record" = TRECORD
getKey "with" = TWITH
getKey "Record" = TBIGRECORD
getKey "Top" = TTOP
getKey _ = TName 

getWhile :: (Char -> Bool) -> String -> String
getWhile _ [] = []
getWhile p (c:cs) = if p c
                    then c : getWhile p cs
                    else []

getRemain :: (Char -> Bool) -> String -> String
getRemain _ [] = []
getRemain p css@(c:cs) = if p c
                         then getRemain p cs
                         else css

---- Parser -------------------------------------------------------------------

type ParserState = String

data ParserResult a =
	  Succ a ParserState
	| Fail

type Parser a = ParserState -> ParserResult a


epsilon :: a -> Parser a
epsilon a = \ps -> Succ a ps


literal :: Terminal -> Parser ()
literal t = \s -> let (nt, v, sr) = getNextToken s in
                            if nt == t
                            then Succ () sr
                            else Fail 

value :: Terminal -> Parser String
value t = \s -> let (nt, v, sr) = getNextToken s in
                            if nt == t
                            then Succ v sr
                            else Fail

(!!!) :: Parser a -> Parser a -> Parser a

p1 !!! p2 = \s -> case p1 s of
			result@(Succ _ _) -> result
			Fail -> p2 s

(&&&) :: Parser a -> Parser b -> Parser b

p1 &&& p2 = \s -> case p1 s of
			Succ v1 s1 -> case p2 s1 of
					result@(Succ _ _) -> result
					Fail -> Fail
			Fail -> Fail

(&&>) :: Parser a -> (a -> Parser b) -> Parser b

p1 &&> p2 = \s -> case p1 s of
			Succ v1 s1 -> case p2 v1 s1 of
					result@(Succ _ _) -> result
					Fail -> Fail
			Fail -> Fail

-- language parsers -----------------------------------------------------------


