{-
	Denotational Semantics
	A dynamically typed language

Denotational Definitions:

Abstract Syntax:

P of Program
C of Command
E of Expression
I of Id
N of Numeral

P ::= C '.'
C ::= C1 ';' C2 | I ':=' E | 'if' E 'then' C1 'else' C2 | 'read' I | 
      'write' E | diverge
E ::= E1 '+' E2 | E1 '=' E2 | '~' E | '(' E ')' | I | N | 'true'

Semantic algebras:

I.	Truth values
	Domain t of Tr = Boolean
	Operations
	  true, false : Tr
	  not : Tr -> Tr

II.	Natural numbers
	Domain n of Nat = natural numbers
	Operations
	  zero, one, ... : Nat
	  plus : Nat x Nat -> Nat
	  equals : Nat x Nat -> Tr

III.	Identifiers
	Domain i of Id = Identifier

IV.	Character strings
	Domain String = the character strings formed from the elements of Char
	(including an 'error' string)
	Operations
	  A, B, C, ..., Z : String
	  empty : String
	  error : String
	  concat : String x String -> String
	  length : String -> Nat
	  substr : String x Nat x Nat -> String

V.	Values that may be stored
	Domain v of Storable-value = Tr + Nat

VI.	Values that expressions may denote
	Domain x of Expressible-value = Storable-value + Errvalue
			where Errvalue = Unit ( () )
	Operations
	  check-expr : (Store -> Expressible-value) x
			(Storable-value -> Store -> Expressible-value) ->
				(Store -> Expressible-value)
	f1 check-expr f2 = \s. cases (f1 s) of
					isStorable-value(v) -> (f2 v s)
					isErrvalue() -> inErrvalue()
					end

VII.	Input buffer
	Domain i of Input = Expressible-value*
	Operations
	  get-value : Input -> (Expressible-value x Input)
	    get-value = \i. null i -> (inErrvalue(),i) else (hd i, tl i)

VIII.	Output buffer
	Domain o of Output = (Storable-value + String)*
	Operations
	  empty : Output
	    empty = nil
	  put-value : Storable-value x Output -> Output
	    put-value = \(v,o).inStorable-value(v) cons o

	  put-message : String x Output -> Output
	    put-message = \(t,o).inString(t) cons o

IX.	Store
	Domain s of Store = Id -> Storable-value
	Operations
	  newstore : Store
	  access : Id -> Store -> Storable-value
	  update : Id -> Storable-value -> Store -> Store

X.	Program State
	Domain a of State = Store x Input x Output

XI.	Post program state
	Domain z of Post-state = OK + Err
			where OK = State
			  and Err = State
	Operations
	  check-result : (Store -> Expressible-value) x 
			   (Storable-value -> State -> Post-state_) ->
				(State -> Post-state_)
	    f check-result g = \(s,i,o). cases (f s) of
				  isStorable-value(v) -> (g v(s,i,o))
				  isErrvalue() -> 
					inErr(s,i,put-message("type error",o))
				  end
	  check-cmd : (State -> Post-state_) x (State -> Post-state_) ->
				(State -> Post-state_)
	    h1 check-cmd h2 = \a. let z = (h1 a) in cases z of
						isOK(s,i,o) -> h2(s,i,o)
						isErr(s,i,o) -> z
						end

Valuation functions:

PP : Program -> Store -> Input -> Post-state_

    PP[C '.'] = \s\i.CC[C](s,i,empty)

CC : Command -> State -> Post-state_

    CC[C1 ';' C2] = CC[C1] check-cmd CC[C2]
    CC[I ':=' E] = EE[E] check-result (\v.\(s,i,o).inOK((update[I] v s),i,o))
    CC['if' E 'then' C1 'else' C2] = EE[E] check-result
		(\v.\(s,i,o).cases v of
			isTr(t) -> (t -> CC[C1] else CC[C2])(s,i,o)
			isNat(n) -> inErr(s,i,put-message("bad test",o))
			end)
    CC['read' I] = \(s,i,o). let(x,i') = get-value(i) in
			cases x of
			 isStorable-value(v) -> inOK((update[I] v s),i',o)
			 isErrvalue() -> inErr(s,i',put-message("bad input",o))
			end
    CC['write' E] = EE[E] check-result (\v.\(s,i,o).inOK(s,i,put-value(v,o)))
    CC['diverge'] = \a._

EE : Expression -> Store -> Expressible-value

    EE[E1 '+' E2] = EE[E1] check-expr
		(\v.cases v of
			isTr(t) -> \s.inErrvalue()
			isNat(n') -> EE[e2] check-expr
				(\v',\s.cases v' of
					isTr(t') -> inErrvalue()
					isNat(n') -> 
					  inStorable-value(inNat(n plus n'))
				end)
		end)
    EE[E1 '=' E2] = "similar to above equation"
    EE['~' E] = EE[E] check-expr
			(\v.\s.cases v of
				isTr(t) -> inStorable-value(inTr(not t))
				isNat(n) -> inErrvalue()
			end)
    EE['(' E ')'] = EE[E]
    EE[I] = \s.inStorable-value(access [I] s)
    EE[N] = \s.inStorable-value(inNat(NN[N]))
    EE['true'] = \s.inStorable-value(inTr(true))

NN : Numeral -> Nat (omitted)

-}

infix `equals`, `plus`

infixr 0 !!!
infixr 1 &&&, &&>

-------------------------------------------------------------------------------
---- Definitions for semantic domains -----------------------------------------
-------------------------------------------------------------------------------

---- Truth values -------------------------------------------------------------

type Tr = Bool

{- true, false, and not are predefined by Gofer:
   true = True
   false = False
   not = not
-}

---- Natural numbers ----------------------------------------------------------

type Nat = Int

plus :: Nat -> Nat -> Nat
n1 `plus` n2 = n1 + n2

equals :: Nat -> Nat -> Tr
n1 `equals` n2 | n1 == n2 = True
               | otherwise = False

---- Identifiers --------------------------------------------------------------

type Id = String

---- Character strings --------------------------------------------------------

{-
  empty = ""
  error = ???
-}

-- concat predefined: concat :: [[a]] -> [a]

-- length predefined: length :: [a] -> Int

substr :: String -> Nat -> Nat -> String
substr s o l = let (head, tail) = splitAt o s in
                       take l tail           

---- Values that may be stored ------------------------------------------------

data StorableValue =
	  TR Tr
	| NAT Nat

---- Values that expressions may denote ---------------------------------------

data ExpressibleValue =
	  Val StorableValue
	| ErrValue

inErrValue :: ExpressibleValue
inErrValue = ErrValue

checkExpression :: (Store -> ExpressibleValue) ->
		   (ExpressibleValue -> Store -> ExpressibleValue) ->
		   (Store -> ExpressibleValue)
f1 `checkExpression` f2 = \s -> case f1 s of
				val@(Val v) -> f2 val s
				ErrValue -> inErrValue

---- Input buffer -------------------------------------------------------------

type Input = [ExpressibleValue]

getValue :: Input -> (ExpressibleValue, Input)
getValue = \i -> if null i then (inErrValue, [])
		           else (head i, tail i)

---- Output buffer ------------------------------------------------------------

data OutputSum =
	  St StorableValue
	| S String

type Output = [OutputSum]

inStorableValue :: StorableValue -> OutputSum
inStorableValue v = St v

inString :: String -> OutputSum
inString t = S t

empty :: Output
empty = []

putValue :: (StorableValue, Output) -> Output
putValue = \(v,o) -> (inStorableValue v) : o

putMessage :: (String, Output) -> Output
putMessage = \(t,o) -> (inString t) : o

---- Store --------------------------------------------------------------------

--type Store = Id -> StorableValue
type Store = [(Id,StorableValue)]

-- newstore :: Store

access :: Id -> Store -> StorableValue
access i ((id,val):s) | id == i = val
                      | otherwise = access i s

update :: Id -> StorableValue -> Store -> Store
update i v [] = [(i,v)]
update i v (entry@(id,val):s) | id == i = (id,v):s
                              | otherwise = entry:(update i v s)

---- Program State ------------------------------------------------------------

type State = (Store,Input,Output)

---- Post program state -------------------------------------------------------

data PostState =
	  OK State
	| Err State
	| Loop

inErr :: State -> PostState
inErr (s,i,o) = Err (s,i,o)

checkResult :: (Store -> ExpressibleValue) -> 
		(StorableValue -> State -> PostState) -> (State -> PostState)
f `checkResult` g = \(s,i,o) -> case f s of
			Val v -> g v (s,i,o)
			ErrValue -> inErr (s, i, (putMessage ("type error",o)))

checkCmd :: (State -> PostState) -> (State -> PostState) ->
		(State -> PostState)
h1 `checkCmd` h2 = \a -> let z = h1 a in case z of
                                       OK (s,i,o) -> h2 (s,i,o)
				       Err (s,i,o) -> z

-------------------------------------------------------------------------------
---- Definitions for abstract syntax ------------------------------------------
-------------------------------------------------------------------------------

data Program =
	  Dot Command

data Command =
	  Semi Command Command
	| Assign Ident Expression
	| If Expression Command Command
	| Read Ident
	| Write Expression
	| Diverge

data Expression =
	  Plus Expression Expression
	| Equal Expression Expression
	| Not Expression
	| Braced Expression
	| I Ident
	| N Numeral
	| ValTrue

data Ident =
	  Identifier Id

data Numeral =
	  Num String

-------------------------------------------------------------------------------
---- Definitions for valuation functions --------------------------------------
-------------------------------------------------------------------------------

--- PP ---

pp :: Program -> Store -> Input -> PostState
pp (Dot c) = \s i -> cc c (s,i,empty)


--- CC ---

cc :: Command -> State -> PostState
cc (Semi c1 c2) = (cc c1) `checkCmd` (cc c2)
cc (Assign id e) = (ee e) `checkResult`
			(\v (s,i,o) -> OK ((update (ii id) v s), i, o))
cc (If e c1 c2) = (ee e) `checkResult` 
			(\v (s,i,o) -> case v of
				TR t -> (if t then (cc c1)
					      else (cc c2))(s,i,o)
                                NAT n -> Err (s,i,(putMessage("bad test",o)))
			 )
cc (Read id) = \(s,i,o) -> let (x,i') = getValue i in
			    case x of
			      Val v -> OK ((update (ii id) v s),i',o)
			      ErrValue->Err (s,i',(putMessage ("bad input",o)))

cc (Write e) = (ee e) `checkResult` (\v (s,i,o) -> OK (s,i,(putValue (v,o))))
cc (Diverge) = \a -> Loop

--- EE ---

ee :: Expression -> Store -> ExpressibleValue
ee (Plus e1 e2) = (ee e1) `checkExpression`
			(\v -> case v of
				(Val (TR t)) -> \s -> ErrValue
				(Val (NAT n)) -> (ee e2) `checkExpression`
				  (\v' s -> case v' of
				    (Val (TR t')) -> ErrValue
				    (Val (NAT n')) -> (Val (NAT (n `plus` n')))
				  )
			)
ee (Equal e1 e2) = (ee e1) `checkExpression`
			(\v -> case v of
				(Val (TR t)) -> \s -> ErrValue
				(Val (NAT n)) -> (ee e2) `checkExpression`
				  (\v' s -> case v' of
				   (Val (TR t')) -> ErrValue
				   (Val (NAT n')) -> (Val (TR (n `equals` n')))
				  )
			)
ee (Not e) = (ee e) `checkExpression` (\v s -> case v of
                                           (Val (TR t)) -> (Val (TR (not t)))
					   (Val _) -> ErrValue )
ee (Braced e) = ee e
ee (I id) = \s -> (Val (access (ii id) s))
ee (N num) = \s -> (Val (NAT (nn num)))
ee (ValTrue) = \s -> (Val (TR True))

--- II ---

ii :: Ident -> Id
ii (Identifier id) = id

--- NN ---

nn :: Numeral -> Nat
nn (Num num) = foldl (\s c -> s * 10 + (ord c - ord '0')) 0 num

-------------------------------------------------------------------------------
---- Definitions for syntax analysis ------------------------------------------
-------------------------------------------------------------------------------

data Terminal = 
	  TSemicolon
	| TDot
	| TAssign
	| TIf
	| TThen
	| TElse
	| TRead
	| TWrite
	| TDiverge
	| TPlus
	| TEqual
	| TNot
	| TLpar
	| TRpar
	| TTrue
	| TName
	| TNumber
	| TEOF

instance Eq Terminal where
	TSemicolon == TSemicolon = True
	TDot == TDot = True
	TAssign == TAssign = True
	TIf == TIf = True
	TThen == TThen = True
	TElse == TElse = True
	TRead == TRead = True
	TWrite == TWrite = True
	TDiverge == TDiverge = True
	TPlus == TPlus = True
	TEqual == TEqual = True
	TNot == TNot = True
	TLpar == TLpar = True
	TRpar == TRpar = True
	TTrue == TTrue = True
	TName == TName = True
	TNumber == TNumber = True
	_ == _ = False

---- Scanner -----------------------------------------------------------------

getNextToken :: String -> ( Terminal, String, String )

getNextToken [] = ( TEOF, "", [] )
getNextToken (';':cs) = ( TSemicolon, "", cs )
getNextToken ('.':cs) = ( TDot, "", cs )
getNextToken (':':'=':cs) = ( TAssign, "", cs )
getNextToken ('+':cs) = ( TPlus, "", cs )
getNextToken ('=':cs) = ( TEqual, "", cs )
getNextToken ('~':cs) = ( TNot, "", cs )
getNextToken ('(':cs) = ( TLpar, "", cs )
getNextToken (')':cs) = ( TRpar, "", cs )

getNextToken (c:cs) | isSpace c = getNextToken cs
                    | isDigit c = ( TNumber, (c:ns), cs' )
                    | isAlpha c = ( (getKey (c:ns')), (c:ns'), cs'' )
                    | otherwise = getNextToken cs
                    where (ns, cs' )  = ( getWhile isDigit cs ,
                                          getRemain isDigit cs )
                          (ns', cs'') = ( getWhile isAlphanum cs,
                                          getRemain isAlphanum cs )

getKey :: String -> Terminal

getKey "if" = TIf
getKey "then" = TThen
getKey "else" = TElse
getKey "read" = TRead
getKey "write" = TWrite
getKey "diverge" = TDiverge
getKey _ = TName 

getWhile :: (Char -> Bool) -> String -> String
getWhile _ [] = []
getWhile p (c:cs) = if p c
                    then c : getWhile p cs
                    else []

getRemain :: (Char -> Bool) -> String -> String
getRemain _ [] = []
getRemain p css@(c:cs) = if p c
                         then getRemain p cs
                         else css

---- Parser -------------------------------------------------------------------

type ParserState = String

data ParserResult a =
	  Succ a ParserState
	| Fail

type Parser a = ParserState -> ParserResult a


epsilon :: a -> Parser a
epsilon a = \ps -> Succ a ps


literal :: Terminal -> Parser ()
literal t = \s -> let (nt, v, sr) = getNextToken s in
                            if nt == t
                            then Succ () sr
                            else Fail 

value :: Terminal -> Parser String
value t = \s -> let (nt, v, sr) = getNextToken s in
                            if nt == t
                            then Succ v sr
                            else Fail

(!!!) :: Parser a -> Parser a -> Parser a

p1 !!! p2 = \s -> case p1 s of
			result@(Succ _ _) -> result
			Fail -> p2 s

(&&&) :: Parser a -> Parser b -> Parser b

p1 &&& p2 = \s -> case p1 s of
			Succ v1 s1 -> case p2 s1 of
					result@(Succ _ _) -> result
					Fail -> Fail
			Fail -> Fail

(&&>) :: Parser a -> (a -> Parser b) -> Parser b

p1 &&> p2 = \s -> case p1 s of
			Succ v1 s1 -> case p2 v1 s1 of
					result@(Succ _ _) -> result
					Fail -> Fail
			Fail -> Fail

---- language parser --------------------------------------------------------

parse_N :: Parser Numeral

parse_N =
	value TNumber &&> (\n -> epsilon (Num n) )

parse_I :: Parser Ident

parse_I =
	value TName &&> (\i -> epsilon (Identifier i ) )

parse_E :: Parser Expression

parse_E =
	parse_E' &&> (\e1 -> literal TPlus &&& parse_E &&>
			(\e2 -> epsilon (Plus e1 e2) ) )

	!!!
	parse_E' &&> (\e1 -> literal TEqual &&& parse_E &&>
			(\e2 -> epsilon (Equal e1 e2) ) )

	!!!

	parse_E'
	
parse_E' :: Parser Expression

parse_E' =
	literal TNot &&& parse_E &&> (\e -> epsilon (Not e) )

	!!!

	literal TLpar &&& parse_E &&> 
		(\e -> literal TRpar &&& epsilon (Braced e) )

	!!!

	parse_I &&> (\id -> epsilon (I id) )

	!!!

	parse_N &&> (\n -> epsilon (N n) )

	!!!

	literal TTrue &&& epsilon (ValTrue)

parse_C :: Parser Command

parse_C =
	parse_C' &&> (\c1 -> literal TSemicolon &&& parse_C &&>
			(\c2 -> epsilon (Semi c1 c2) ) )

	!!!

	parse_C'

parse_C' :: Parser Command

parse_C' =
	parse_I &&> (\i -> literal TAssign &&& parse_E &&>
			(\e -> epsilon(Assign i e) ) )

	!!!

	literal TIf &&& parse_E &&>
		(\e -> literal TThen &&& parse_C &&>
		 (\c1 -> literal TElse &&& parse_C &&>
		  (\c2 -> epsilon (If e c1 c2) ) ) )

	!!!

	literal TRead &&& parse_I &&> (\i -> epsilon (Read i) )

	!!!

	literal TWrite &&& parse_E &&> (\e -> epsilon (Write e) )

	!!!

	literal TDiverge &&& epsilon (Diverge)

parse_P :: Parser Program

parse_P =
	parse_C &&> (\c -> literal TDot &&& epsilon (Dot c) )

------------------------------------------------------------------------------
---- Interpreter -------------------------------------------------------------
------------------------------------------------------------------------------

interpreter :: String -> PostState

interpreter = \s -> case parse_P s of
			Succ p _ -> pp p [] []
			Fail -> Err ([],[],[])


